# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: odsbox/proto/ods_notification.proto
# Protobuf Python Version: 5.26.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n#odsbox/proto/ods_notification.proto\x12\x10ods.notification\"\xa8\x02\n\x13RegistrationRequest\x12H\n\x04mode\x18\x01 \x01(\x0e\x32:.ods.notification.RegistrationRequest.NotificationModeEnum\x12\x42\n\x07\x65ntries\x18\x02 \x03(\x0b\x32\x31.ods.notification.RegistrationRequest.TypeElement\x1aQ\n\x0bTypeElement\x12\x34\n\x04type\x18\x01 \x01(\x0e\x32&.ods.notification.NotificationTypeEnum\x12\x0c\n\x04\x61ids\x18\x02 \x03(\x03\"0\n\x14NotificationModeEnum\x12\x0b\n\x07NM_POOL\x10\x00\x12\x0b\n\x07NM_PUSH\x10\x01\"\x91\x01\n\x0cNotification\x12\x0c\n\x04uuid\x18\x01 \x01(\t\x12\x34\n\x04type\x18\x02 \x01(\x0e\x32&.ods.notification.NotificationTypeEnum\x12\x0b\n\x03\x61id\x18\x03 \x01(\x03\x12\x0c\n\x04iids\x18\x04 \x03(\x03\x12\x11\n\ttimestamp\x18\x05 \x01(\t\x12\x0f\n\x07user_id\x18\x06 \x01(\x03\"B\n\x10NotificationPool\x12.\n\x06\x65vents\x18\x01 \x03(\x0b\x32\x1e.ods.notification.Notification*k\n\x14NotificationTypeEnum\x12\n\n\x06NT_ALL\x10\x00\x12\n\n\x06NT_NEW\x10\x01\x12\r\n\tNT_MODIFY\x10\x02\x12\r\n\tNT_DELETE\x10\x03\x12\x0c\n\x08NT_MODEL\x10\x04\x12\x0f\n\x0bNT_SECURITY\x10\x05\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'odsbox.proto.ods_notification_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_NOTIFICATIONTYPEENUM']._serialized_start=572
  _globals['_NOTIFICATIONTYPEENUM']._serialized_end=679
  _globals['_REGISTRATIONREQUEST']._serialized_start=58
  _globals['_REGISTRATIONREQUEST']._serialized_end=354
  _globals['_REGISTRATIONREQUEST_TYPEELEMENT']._serialized_start=223
  _globals['_REGISTRATIONREQUEST_TYPEELEMENT']._serialized_end=304
  _globals['_REGISTRATIONREQUEST_NOTIFICATIONMODEENUM']._serialized_start=306
  _globals['_REGISTRATIONREQUEST_NOTIFICATIONMODEENUM']._serialized_end=354
  _globals['_NOTIFICATION']._serialized_start=357
  _globals['_NOTIFICATION']._serialized_end=502
  _globals['_NOTIFICATIONPOOL']._serialized_start=504
  _globals['_NOTIFICATIONPOOL']._serialized_end=570
# @@protoc_insertion_point(module_scope)
